# Specification of the Web API for the Video-AG Website (v0.54).

## Introduction

In this document we use a lot of URLs. Parameters which need to be replaced are put in `{}`. Anything inside and the
braces need to be replaced with the actual argument. These are called _URL Parameters_.

The Base-URL is `/api/{version}`, where `{version}` must be a version id from below. All URLs in this document are
relative to this Base-URL (Unless explicitly
stated otherwise).

The versions are:

| Version | id | Status         |
|---------|----|----------------|
| 0       | v0 | In Development |

HTTP Bodies, if present, are always JSON objects unless explicitly stated otherwise. We use tables to represent the json
object structure:

| Field     | Type    | Notes                    |
|-----------|---------|--------------------------|
| id        | int     |                          |
| something | ?string | We don't know what it is |

Example json object:
<pre>
{
    "id": 42,
    "something": "asdfghjkl"
}
</pre>

The _Field_ is the name of the JSON Key. _Type_ is the type of data. The type may have some additional indicators:

* A question mark indicates that the value is optional.
* An exclamation mark indicates that the value may only be present if the user is a moderator.
* A plus indicates that the field is [directly modifiable](#directly-modifiable-fields)

Some routes, marked as such, require a CSRF token. This token is set as a cookie `csrf_token` by
`POST /authentication/fsmpi` or `POST /authentication/status` and must be included as the `X-Csrf-Token` HTTP Header in
such requests.

On success, status code `200 OK` is always returned unless otherwise noted. Response Body is always present. For
status codes `4xx` and `5xx` the response contains an [api_error](#api_error) (May not be present if the error
happened in a proxy, etc.).

The changelog can be found [at the end](#changelog).

## Routes

### Site status

The current site status can be retrieved with `GET /status`. It can be in three states

* `available`: The site is running and all routes can be used
* `readonly`: The site cannot update any data. You can only call routes for which is specified:
  * 'This may be used while the site is [read-only](#site-status).'
* `unavailable`: The site is unavailable and the only route you can call is `GET /status`

---

#### `GET /status`

This may be used even if the site is read-only or disabled.

###### Response:

| Field           | Type                                    | Notes                                                   |
|-----------------|-----------------------------------------|---------------------------------------------------------|
| status          | id_string                               | possible values: `available`, `readonly`, `unavailable` |
| is_debug        | boolean                                 |                                                         |
| git_commit_hash | ?string                                 |                                                         |
| server_name     | string                                  | Name of the server which is answering the request       |
| announcements   | array of [announcement](#announcement)  |                                                         |

<details>
<summary>Example: <code>GET /status</code></summary>
<pre>
{
    "status": "running",
    "is_debug": true,
    "server_name": "dev",
    "announcements": [
        {
            "type": "info",
            "visibility": "only_main_page",
            "text": "We are working on a new website!"
        },
        {
            "type": "warning",
            "visibility": "all_pages",
            "text": "There will be maintenance downtime"
        }
    ]
}
</pre>
</details>

---

#### `GET /is_running`

This may be used even if the site is read-only or disabled.

###### Response:

Returns `200 OK` if the site is ready to answer requests, otherwise `503 SERVICE UNAVAILABLE`. The status may still be
`unavailable` if this return `200 OK`!

---

### Miscellaneous

#### `GET /homepage`

This may be used while the site is [read-only](#site-status).

###### Response:

A [homepage](#homepage) object.

<details>
<summary>Example: <code>GET /homepage</code></summary>
<pre>
{
    "featured": [
        {
            "type": "plain",
            "title": "Video AG",
            "text": "Wir machen Vorlesungsvideos, damit du dir deine Vorlesungen angucken kannst, wann, wo und so oft du willst ;)"
        }
    ],
    "upcoming_lectures": [],
    "new_lectures": [],
    "courses_context": {}
}
</pre>
</details>

---

#### `GET /lecture_import/rwth_online/{rwth_course_id}`

| URL Parameter   | Type | Notes                                     |
|-----------------|------|-------------------------------------------|
| rwth_course_id  | int  | ID of the course to fetch from RWTHonline |

Only for moderators. Requires CSRF token

###### Response:

Queries all events for the specified course from RWTHonline in ICAL format.

Content Type: `text/calendar`

---

### Courses and Lectures

The results of these routes depend on whether the user is a moderator. Objects which are not visible or not listed are only
included for moderators.

#### `GET /courses`

This may be used while the site is [read-only](#site-status).

###### Response:

| Field   | Type                     | Notes                                                           |
|---------|--------------------------|-----------------------------------------------------------------|
| courses | map of [course](#course) | Key is course id. The courses **do not** include their lectures |

<details>
<summary>Example: <code>GET /course</code></summary>
<pre>
{
    "courses": {
        "1068": {
            "id": 1068,
            "id_string": "23ws-buk",
            "full_name": "Berechenbarkeit und Komplexität",
            "short_name": "BuK",
            "semester": "2023ws",
            "organizer": "Prof. Rossmanith",
            "topic": "Informatik",
            "description": "",
            "show_chapters_on_course": true
           "default_authentication_methods": ["public"],
        },
        "1063": {
            "id": 1063,
            "id_string": "23ws-malo2",
            "full_name": "Mathematische Logik II",
            "short_name": "MaLo2",
            "semester": "2023ws",
            "organizer": "Prof. Grädel",
            "topic": "Informatik, Mathematik",
            "description": "",
            "show_chapters_on_course": true
           "default_authentication_methods": ["rwth"],
        }
    }
}
</pre>
</details>

---

#### `GET /course/{course_id}?include_lectures={include_lectures}`

| URL Parameter | Type             | Notes                                    |
|---------------|------------------|------------------------------------------|
| course_id     | int or id_string | Either unique int id or unique id_string |

| Query parameter  | Type     | Notes                                 | Default |
|------------------|----------|---------------------------------------|---------|
| include_lectures | ?boolean | If `true`, `lectures` will be present | false   |

This may be used while the site is [read-only](#site-status).

###### Response:

A [course](#course) object.

<details>
<summary>Example: <code>GET /course/1068?include_lectures=true</code></summary>
<pre>
{
    "id": 1068,
    "id_string": "23ws-buk",
    "full_name": "Berechenbarkeit und Komplexität",
    "short_name": "BuK",
    "semester": "2023ws",
    "organizer": "Prof. Rossmanith",
    "topic": "Informatik",
    "description": "",
    "show_chapters_on_course": true,
    "default_authentication_methods": ["public"],
    "lectures": [
        {
            "id": 17256,
            "course_id": 1068,
            "title": "Turing-Maschinen I",
            "speaker": "Prof. Rossmanith",
            "location": "AM",
            "time": "2023-10-13T14:30:00",
            "duration": 90,
            "description": "Die Kameraaufnahme ist aufgrund einer korrumpierten SD-Karte verloren gegangen, daher nur Slides und Audio",
            "thumbnail_url": "https://video.fsmpi.rwth-aachen.de/files/thumbnail/l_17256.jpg",
            "no_recording": false,
            "live_stream_planned": false
            "chapters": [
                {
                    "start_time": 1401,
                    "name": "Funktionsweise von Turing-Maschinen"
                },
                {
                    "start_time": 2399,
                    "name": "Beispiele von Turing-Maschinen"
                },
                {
                    "start_time": 3413,
                    "name": "Programmierung der Turing-Maschine"
                },
                {
                    "start_time": 4194,
                    "name": "Konfigurationen"
                }
            ],
            "media_sources": [
                {
                    "quality": {
                        "name": "1080p",
                        "resolution": "1920x1080",
                        "aspect_ratio": "16:9",
                        "priority": 5
                    },
                    "size": 457074278,
                    "comment": "",
                    "player_url": "https://video.fsmpi.rwth-aachen.de/files/pub/23ws-buk/23ws-buk-231013-1080p.mp4",
                    "download_url": "https://video.fsmpi.rwth-aachen.de/files/pub/23ws-buk/23ws-buk-231013-1080p.mp4"
                }
            ],
            "authentication_methods": ["public"]
        }
    ]
}
</pre>
</details>

---

#### `GET /lecture/{lecture_id}`

| URL Parameter | Type | Notes |
|---------------|------|-------|
| lecture_id    | int  |       |

This may be used while the site is [read-only](#site-status).

###### Response:

| Field   | Type                | Notes                                            |
|---------|---------------------|--------------------------------------------------|
| course  | [course](#course)   | The course **does not** include all its lectures |
| lecture | [lecture](#lecture) | Includes chapters and media_sources              |

<details>
<summary>Example: <code>GET /lecture/17256</code></summary>
<pre>
{
    "course": {
        "id": 1068,
        "id_string": "23ws-buk",
        "full_name": "Berechenbarkeit und Komplexität",
        "short_name": "BuK",
        "semester": "2023ws",
        "organizer": "Prof. Rossmanith",
        "topic": "Informatik",
        "description": "",
        "show_chapters_on_course": true,
        "default_authentication_methods": ["public"]
    },
    "lecture": {
        "id": 17256,
        "course_id": 1068,
        "title": "Turing-Maschinen I",
        "speaker": "Prof. Rossmanith",
        "location": "AM",
        "time": "2023-10-13T14:30:00",
        "duration": 90,
        "description": "Die Kameraaufnahme ist aufgrund einer korrumpierten SD-Karte verloren gegangen, daher nur Slides und Audio",
        "thumbnail_url": "https://video.fsmpi.rwth-aachen.de/files/thumbnail/l_17256.jpg",
        "no_recording": false,
        "livestream_planned": false,
        "chapters": [
            {
                "start_time": 1401,
                "name": "Funktionsweise von Turing-Maschinen"
            },
            {
                "start_time": 2399,
                "name": "Beispiele von Turing-Maschinen"
            },
            {
                "start_time": 3413,
                "name": "Programmierung der Turing-Maschine"
            },
            {
                "start_time": 4194,
                "name": "Konfigurationen"
            }
        ],
        "media_sources": [
            {
                "quality": {
                    "name": "1080p",
                    "resolution": "1920x1080",
                    "aspect_ratio": "16:9",
                    "priority": 5
                },
                "size": 457074278,
                "comment": "",
                "player_url": "https://video.fsmpi.rwth-aachen.de/files/pub/23ws-buk/23ws-buk-231013-1080p.mp4",
                "download_url": "https://video.fsmpi.rwth-aachen.de/files/pub/23ws-buk/23ws-buk-231013-1080p.mp4"
            }
        ],
        "authentication_methods": ["public"]
    }
}
</pre>
</details>

---

#### `GET /search?q={search_term}`

| URL Parameter | Type   | Notes                                                                                     |
|---------------|--------|-------------------------------------------------------------------------------------------|
| search_term   | string | The term to search for. May not be empty (only whitespace) or longer than 1024 characters |

This may be used while the site is [read-only](#site-status).

###### Response:

| Field           | Type                         | Notes                                                                                                                                   |
|-----------------|------------------------------|-----------------------------------------------------------------------------------------------------------------------------------------|
| courses         | array of int                 | Contains the ids of found courses                                                                                                       |
| lectures        | array of [lecture](#lecture) | Includes chapters and media_sources                                                                                                     |
| courses_context | map of [course](#course)     | Key is id of course. This contains all courses which are referenced in _courses_ or _lectures_. They **do not** include their lectures. |

---

#### `PUT /lecture/{lecture_id}/chapter_suggestion`

| URL Parameter | Type | Notes |
|---------------|------|-------|
| lecture_id    | int  |       |

| Field      | Type      | Notes |
|------------|-----------|-------|
| start_time | timestamp |       |
| name       | string    |       |

###### Response:

`201 CREATED` if suggestion was saved or an [api_error](#api_error).

<details>
<summary>Example: <code>PUT /lecture/17256/chapter_suggestion</code></summary>
<pre>
{
    "start_time": "60",
    "name": "Wiederholung"
}
</pre>
Response: `200 OK`
</details>

---

### Authentication

#### `POST /authentication/password`

| Field      | Type   | Notes          |
|------------|--------|----------------|
| username   | string |                |
| password   | string |                |
| lecture_id | int    | Id of lecture  |

If the specified lecture does not exist or does not have password authentication, an [api error](#api_error) is returned.

This may be used while the site is [read-only](#site-status).

###### Response:

`200 OK` if the username and password are correct for the specified lecture. Otherwise, an [api error](#api_error) (Usually `authentication_failed`).

The username and password are saved, so that different lectures with the same login data are authenticated as well.

---

#### `POST /authentication/fsmpi`

| Field      | Type   | Notes                                                 |
|------------|--------|-------------------------------------------------------|
| username   | string |                                                       |
| password   | string |                                                       |
| lecture_id | ?int   | Id of lecture for which to get authentication status. |

If the specified lecture does not exist, a (random) result or an [api error](#api_error) may be returned.

This may be used while the site is [read-only](#site-status).

###### Response:

`200 OK` if login was successful. Otherwise, an [api error](#api_error) (Usually `authentication_failed`).

| Field                    | Type       | Notes                                              |
|--------------------------|------------|----------------------------------------------------|
| user                     | ?user_info | Only present if the user has moderator access      |
| is_lecture_authenticated | ?boolean   | Only present if lecture_id was present in request. |

---

#### `POST /authentication/start_oauth`

| Field | Type   | Notes                             |
|-------|--------|-----------------------------------|
| type  | string | Possible values: `rwth`, `moodle` |

This may be used while the site is [read-only](#site-status).

###### Response:

`202 ACCEPTED` with body:

| Field            | Type | Notes                                          |
|------------------|------|------------------------------------------------|
| verification_url | url  | URL at which user must authenticate themselves |

After the user has authenticated themselves, the authentication needs to be finished by requesting the
status: `/authentication/status`.

---

#### `POST /authentication/status`

| Field      | Type | Notes                                                |
|------------|------|------------------------------------------------------|
| lecture_id | ?int | Id of lecture for which to get authentication status |

If the specified lecture does not exist, a (random) result or an [api error](#api_error) may be returned.

This may be used while the site is [read-only](#site-status).

###### Response:

Also finishes the last authentication started with `/authentication/start`. If that authentication is not yet complete,
`in_progress_authentication` will be set (Such an authentication cannot 'fail', only be successful or timeout).

| Field                      | Type                      | Notes                                                  |
|----------------------------|---------------------------|--------------------------------------------------------|
| authenticated_methods      | array of string           | Possible values: `fsmpi`, `rwth`, `moodle`, `password` |
| in_progress_authentication | ?string                   | Possible values: `rwth`, `moodle`                      |
| is_lecture_authenticated   | ?boolean                  | Only present if lecture_id was present in request.     |
| user_info                  | \![user_info](#user_info) |                                                        |

---

#### `POST /authentication/moderator_logout`

This may be used while the site is [read-only](#site-status).

###### Response:

Only logs moderators out.

`200 OK` if now logged out

---

### Feedback

#### `PUT /feedback/new`

| Field      | Type    | Notes                                                                                                            |
|------------|---------|------------------------------------------------------------------------------------------------------------------|
| email      | ?string | May not be longer than 128 characters. Must match `\S+@\S+\.\S+` (More sophisticated checks are allowed, though) |
| text       | string  | May not be longer than 16384 characters                                                                          |

###### Response:

`201 CREATED` if the feedback was saved. Otherwise, an [api error](#api_error).

---

#### `GET /feedback?entries_per_page={entries_per_page}&page={page}`

| Query Parameter  | Type       | Notes                                                                                                |
|------------------|------------|------------------------------------------------------------------------------------------------------|
| entries_per_page | ?int       | Must be between 10 and 500                                                                           |
| page             | ?int       | Zero-indexed. Must not be negative. If this is bigger than the page count, an empty page is returned |

This may be used while the site is [read-only](#site-status).

This route is only for moderators.

###### Response:

| Field      | Type                                       | Notes |
|------------|--------------------------------------------|-------|
| page_count | int                                        |       |
| page       | array of [feedback_entry](#feedback_entry) |       |

---

### Objects

These routes allow moderators to change fields of objects. For most objects this is done indirectly. The API
provides a configuration which specifies all the fields with their types and current values. These properties can
be dynamically displayed. Changes can then be made with an API request.

#### Objects with configuration overview

| Object       | Parent object   | `GET/PATCH .../configuration` | `GET/PUT .../new/...` | `DELETE ...` | Notes                                              |
|--------------|-----------------|-------------------------------|-----------------------|--------------|----------------------------------------------------|
| announcement |                 | ✓                             | ✓                     | ✓            |                                                    |
| featured     |                 | ✓                             | ✓                     | ✓            |                                                    |
| course       |                 | ✓                             | ✓                     | ✓            |                                                    |
| lecture      | course          | ✓                             | ✓                     | ✓            |                                                    |
| chapter      | lecture         | ✓                             | ✓                     | ✓            |                                                    |
| media_source | lecture         | ✓                             |                       | ✓            |                                                    |

These routes are **only** accessible by moderators

#### Directly Modifiable Fields

Modification for some fields, which directly represent a json value in an object, may also be modified directly.
These may be provided in a configuration as well, but they are guaranteed to exist and modifiable with the ids and types
specified in this document. These fields are marked with `+` in their respective objects. Note that some additional
restrictions may apply for which the configuration must be queried. Most notable, strings usually have a maximum length.

---

#### `GET /object_management/{object_type}/{object_id}/configuration`

| URL Parameter | Type      | Notes                                                                                                                   |
|---------------|-----------|-------------------------------------------------------------------------------------------------------------------------|
| object_type   | id_string | Name of object (as used in this document) which has a configuration (See [above](#Objects-with-configuration-overview)) |
| object_id     | int       |                                                                                                                         |

This may be used while the site is [read-only](#site-status).

###### Response:

Returns the configuration for the object of the specified type and id or an [api error](#api_error) if it does not exist.

<details>
<summary>Example: <code>GET /course/1068/configuration</code></summary>
<pre>
{
    "permissions": [
        {
            "id": 42,
            "type": "password"
            "username": "mustermann",
            "password": "this.is.safe"
        }
    ],
    "fields": [
        {
            "description": {
                "id": "title",
                "type": "string",
                "min_length": 3,
                "max_length": 100
            },
            "value": "Berechenbarkeit und Komplexität"
        },
        {
            "description": {
                "id": "downloadable",
                "type": "boolean"
            },
            "value": true
        },
        {
            "description": {
                "id": "id_string",
                "type": "id_string"
            },
            "value": "23ws-buk"
        }
    ]
}
</pre>
</details>

---

#### `PATCH /object_management/{object_type}/{object_id}/configuration`

| URL Parameter | Type      | Notes                                                                                                                   |
|---------------|-----------|-------------------------------------------------------------------------------------------------------------------------|
| object_type   | id_string | Name of object (as used in this document) which has a configuration (See [above](#Objects-with-configuration-overview)) |
| object_id     | int       |                                                                                                                         |

| Field                   | Type | Notes                                                                                                                                                     |
|-------------------------|------|-----------------------------------------------------------------------------------------------------------------------------------------------------------|
| updates                 | map  | Key is `id_string` of field (Provided by the configuration or of a [directly modifiable field](#Directly-Modifiable-Fields)). Value has type of the field |
| expected_current_values | map  | Same format as `updates`. An error is thrown if the current value is not the value specified here for some field                                          |

Requires CSRF token

###### Response:

Updates the specified fields or returns an [api error](#api_error) if the update failed.

<details>
<summary>Example: <code>PATCH /course/10650/configuration</code></summary>
<pre>
{
    "updates": [
        "title": "Vorlesung XY",
        "downloadable": false
    ]
}
</pre>

Response: <code>200 OK</code>
</details>

---

#### `POST /object_management/modify_many`

| Field   | Type                                      | Notes                                                         |
|---------|-------------------------------------------|---------------------------------------------------------------|
| objects | array of object_update as described below | An object may not occur multiple times. Maximum is 32 objects |

object_update:

| Field   | Type      | Notes                                                                                                                   |
|---------|-----------|-------------------------------------------------------------------------------------------------------------------------|
| type    | id_string | Name of object (as used in this document) which has a configuration (See [above](#Objects-with-configuration-overview)) |
| id      | int       | Id of the object to update                                                                                              |
| updates | map       | Same format as `updates` in `PATCH /{object_type}/{object_id}/configuration`                                            |

Requires CSRF token

###### Response:

Updates the specified objects or returns an [api error](#api_error) if the update failed.

<details>
<summary>Example: <code>POST /objects/modify_many</code></summary>
<pre>
{
    "objects": [
        {
            "type": "course",
            "id": 10650,
            "updates": {
                "title": "Vorlesung XY",
                "downloadable": false
            }
        },
        {
            "type": "lecture",
            "id": 89457,
            "updates": {
                "title": "Einführung",
                "is_visible": true
            }
        }
    ]
}
</pre>

Response: <code>200 OK</code>
</details>

---

#### `GET /object_management/{object_type}/new/configuration`

| URL Parameter | Type      | Notes                                                                                                                   |
|---------------|-----------|-------------------------------------------------------------------------------------------------------------------------|
| object_type   | id_string | Name of object (as used in this document) which has a configuration (See [above](#Objects-with-configuration-overview)) |

This may be used while the site is [read-only](#site-status).

###### Response:

Returns the fields which may need to be configured to create a new object. For the creation all returned fields **may** be specified. All fields which do not have a default value **must** be specified.

| Field          | Type                                             | Notes                                                                                                                                                                    |
|----------------|--------------------------------------------------|--------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| fields         | array of [field_description](#field_description) |                                                                                                                                                                          |
| variant_fields | ?map of array as described as above              | If present, a variant for the object must be chosen. Possible variants are keys of map (type `id_string`). Additional fields for that variant are specified in the array |

<details>
<summary>Example: <code>GET /featured/new/configuration</code></summary>
<pre>
{
    "fields": [
        "description": {
            "id": "title",
            "type": "string"
        },
        "description": {
            "id": "text",
            "type": "string"
        },
        "description": {
            "id": "is_visible",
            "type": "string",
            "default_value": false
        }
    ],
    "variant_fields": {
        "text": [],
        "image": [
            {
                "id": "image_url",
                "type": "string"
            }
        ],
        "lecture": [
            {
                "id": "lecture_id",
                "type": "lecture_id"
            }
        ],
        "course_list": []
    }
}
</pre>
</details>

---

#### `PUT /object_management/{object_type}/new`

| URL Parameter | Type      | Notes                                                                                                                   |
|---------------|-----------|-------------------------------------------------------------------------------------------------------------------------|
| object_type   | id_string | Name of object (as used in this document) which has a configuration (See [above](#Objects-with-configuration-overview)) |

| Field       | Type       | Notes                                                                                                                                             |
|-------------|------------|---------------------------------------------------------------------------------------------------------------------------------------------------|
| parent_type | ?id_string | Type of parent object. Must be present if object can have parent. See [Objects with configuration overview](#Objects-with-configuration-overview) |
| parent_id   | ?int       | Id of the parent object. Must be present if, and only if, `parent_type` is present.                                                               |
| values      | map        | Key is `id_string` of field (Provided by the creation configuration; for the specified object type if one is set). Value has type of the field    |
| variant     | ?id_string | Must be present if `GET /{object_type}/new/configuration` included `variant_fields`. Value must be a key of that map                              |

Requires CSRF token.

For some objects it is guaranteed that they have no variant and that the following fields are sufficient to create
these objects (No other creation fields apart from these, have no default value):

For `lecture` (With types as specified in the `lecture` object): 
* `title`
* `location`
* `time`
* `duration`

For `chapter` (With types as specified in the `chapter` object): 
* `start_time`
* `name`
* `is_visible`

###### Response:

`201 CREATED` if the object was created successfully, otherwise an [api error](#api_error). 

| Field | Type | Notes                          |
|-------|------|--------------------------------|
| id    | int  | Id of the newly created object |

<details>
<summary>Example: <code>PUT /featured/new</code></summary>
<pre>
{
    "values": [
        "title": "Sehenswerte Vorlesung",
        "text": "Das müsst ihr euch unbedingt anschauen!",
        "is_visible": true,
        "lecture_id": 42
    ],
    "variant": "lecture"
}
</pre>
Response: <code>201 CREATED</code>
<pre>
{
    "id": 240
}
</pre>
</details>

---

#### `DELETE /object_management/{object_type}/{object_id}`

| URL Parameter | Type      | Notes                                                                                                              |
|---------------|-----------|--------------------------------------------------------------------------------------------------------------------|
| object_type   | id_string | Name of object (as used in this document) which may be deleted (See [above](#Objects-with-configuration-overview)) |
| object_id     | int       |                                                                                                                    |

Requires CSRF token

###### Response:

Deletes specified object or returns an [api error](#api_error) if the deletion failed (Does not exist, already deleted).

---

#### `POST /object_management/{object_type}/{object_id}/resurrect`

| URL Parameter | Type      | Notes                                                                                                              |
|---------------|-----------|--------------------------------------------------------------------------------------------------------------------|
| object_type   | id_string | Name of object (as used in this document) which may be deleted (See [above](#Objects-with-configuration-overview)) |
| object_id     | int       |                                                                                                                    |

Requires CSRF token

###### Response:

Restores the specified object, which must have been deleted previously, or returns an [api error](#api_error) if it
could not be restored (Does not exist, not deleted).

---

#### `GET /object_management/changelog?object_type={object_type}&object_id={object_id}&field={field}&entries_per_page={entries_per_page}&page={page}`

| Query Parameter  | Type       | Notes                                                                                                |
|------------------|------------|------------------------------------------------------------------------------------------------------|
| entries_per_page | ?int       | Must be between 10 and 500                                                                           |
| page             | ?int       | Zero-indexed. Must not be negative. If this is bigger than the page count, an empty page is returned |
| object_type      | ?id_string | Only return entries for objects of this type                                                         |
| object_id        | ?int       | Only return entries for objects with this id                                                         |
| field            | ?id_string | Only return entries for this field. `object_type` must be set                                        |

This may be used while the site is [read-only](#site-status).

###### Response:

| Field      | Type                                         | Notes |
|------------|----------------------------------------------|-------|
| page_count | int                                          |       |
| page       | array of [changelog_entry](#changelog_entry) |       |

---

### Sorter

The following routes are only for moderators.

#### `GET /sorter/log`

This may be used while the site is [read-only](#site-status).

###### Response:

| Field     | Type                                                       | Notes |
|-----------|------------------------------------------------------------|-------|
| log       | array of [sorter_log_entry](#sorter_log_entry)             |       |
| error_log | array of [sorter_error_log_entry](#sorter_error_log_entry) |       |

---

#### `DELETE /sorter/error_log/{log_id}`

| URL Parameter | Type | Notes |
|---------------|------|-------|
| log_id        | int  |       |

Requires CSRF token.

###### Response:

Deletes specified error log entry or returns an [api error](#api_error) if the deletion failed.

---

### Users

The following routes are only for moderators.

#### `GET /users`

This may be used while the site is [read-only](#site-status).

###### Response:

Returns a list of all users (moderators)

| Field                      | Type                             | Notes |
|----------------------------|----------------------------------|-------|
| users                      | array of [user_info](#user_info) |       |

---

#### `GET /user/me/settings`

This may be used while the site is [read-only](#site-status).

###### Response:

A [settings](#settings) object

---

#### `PATCH /user/me/settings`

| Field   | Type | Notes                                                                                                                                       |
|---------|------|---------------------------------------------------------------------------------------------------------------------------------------------|
| updates | map  | Key is id of setting_entry from `GET /user/me/settings` to change (All types except `subsection`). Value is new value of corresponding type |


Requires CSRF token.

###### Response:

Updates the specified settings or returns an [api_error](#api_error) if the update failed.

---

### Jobs

The following routes are only for moderators.

#### `GET /jobs?entries_per_page={entries_per_page}&page={page}&type={type}&worker={worker_id}&state={state}`

| Query Parameter  | Type       | Notes                                                                                                |
|------------------|------------|------------------------------------------------------------------------------------------------------|
| entries_per_page | ?int       | Must be between 10 and 500                                                                           |
| page             | ?int       | Zero-indexed. Must not be negative. If this is bigger than the page count, an empty page is returned |
| type             | ?job_type  | Only get jobs of this type                                                                           |
| worker_id        | ?string    | Only get jobs for this worker. May be `none` to only get jobs assigned to no worker                  |
| state            | ?job_state | Only get jobs in this state                                                                          |

This may be used while the site is [read-only](#site-status).

###### Response:

| Field      | Type                       | Notes |
|------------|----------------------------|-------|
| page_count | int                        |       |
| page       | array of [job](#job)       |       |
| workers    | array of [worker](#worker) |       |

---

#### `DELETE /job/{id}`

| Query Parameter | Type | Notes                             |
|-----------------|------|-----------------------------------|
| id              | int  | Id of **ready** or **failed** job |

Requires CSRF token.

###### Response:

Deletes the specified job.

---

#### `POST /job/{id}/action`

| Query Parameter | Type | Notes |
|-----------------|------|-------|
| id              | int  |       |

| Field  | Type      | Notes                                                                           |
|--------|-----------|---------------------------------------------------------------------------------|
| action | id_string | For possible actions see below. Note that the job has to be in a specific state |

| Action    | For jobs in state(s) | Notes |
|-----------|----------------------|-------|
| `restart` | failed               |       |
| `cancel`  | running              |       |
| `copy`    | finished, deleted    |       |

Requires CSRF token.

###### Response:

Executes the specified action.

---

## Resources

### Types

| Type                     | Actual data type | May be type of field | Notes                                                                                                                                                                                  |
|--------------------------|------------------|----------------------|----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| int                      |                  | ✓                    | signed 32-bit integer                                                                                                                                                                  |
| long                     |                  | ✓                    | signed 64-bit integer                                                                                                                                                                  |
| boolean                  |                  | ✓                    |                                                                                                                                                                                        |
| string                   |                  | ✓                    |                                                                                                                                                                                        |
| id_string                | string           |                      | must match `(?=.*[a-z_-])[a-z0-9_-]{1,100}`                                                                                                                                            |
| semester_string          | string           | ✓                    | must match `([0-9]{4}(ws\|ss)\|none)`                                                                                                                                                  |
| duration                 | int              |                      | \>= 0, in minutes                                                                                                                                                                      |
| datetime                 | string           | ✓                    | must be in format `YYYY-MM-ddTHH:mm:ss` (Year, Month, day, Hour, minute, second) (Note that this is a local date time according to [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601)) |
| timestamp                | int              |                      | \>= 0, in milliseconds                                                                                                                                                                 |
| url                      | string           | ✓                    | must be an [URL](https://en.wikipedia.org/wiki/URL)                                                                                                                                    |
| data_size                | long             |                      | in bytes                                                                                                                                                                               |
| authentication_methods   | array of string  |                      | possible values for string: `public` (Everyone has access, array will contain no other values), `rwth`, `moodle`, `fsmpi`, `password`; Empty array indicates no access                 |
| legacy_announcement_type | string           | ✓                    | possible values: `info_main_page`, `info_all_pages`, `warning_all_pages`, `important_all_pages_embed`. Will be removed after transition                                                |
| announcement_type        | string           | ✓                    | possible values: `info`, `warning`, `important`                                                                                                                                        |
| announcement_visibility  | string           | ✓                    | possible values: `only_main_page`, `all_pages`, `all_pages_and_embed`                                                                                                                  |
| featured_type            | string           | ✓                    | possible values: `plain`, `image`, `lecture`, `course_list`                                                                                                                            |
| view_permissions_type    | string           | ✓                    | possible values: `public`, `private`, `authentication`, `inherit`                                                                                                                      |
| course_filter_criteria   | string           | ✓                    | possible values: `title`, `semester`, `organizer`, `topic`                                                                                                                             |
| lecture_id               | int              | ✓                    | Existence must be checked                                                                                                                                                              |
| user_id                  | int              | ✓                    | Existence must be checked                                                                                                                                                              |
| user_id_list             | array of user_id | ✓                    |                                                                                                                                                                                        |
| course_id_string         | string           | ✓                    | Must match `[a-zA-Z0-9_-]{1,100}`, must not be used by other course, must **not** be one of: `courses`, `faq`, `imprint`, `licenses`, `search`, `site`, `internal`                     |
| job_state                | id_string        |                      | possible values: `ready`, `running`, `finished`, `failed`, `deleted`                                                                                                                   |
| job_type                 | id_string        |                      | possible values: `probe`, `probe-raw`, `remux`, `thumbnail`, `transcode`, `publish_video`, `simple_live_transcode`, `complex_live_transcode`, `live_forward`                           |

Additionally, the following objects may appear as the type of some field:

* `view_permissions`

### Objects

#### announcement

| Field                | Type                    | Notes                                                                                                                   |
|----------------------|-------------------------|-------------------------------------------------------------------------------------------------------------------------|
| id                   | ?int                    | If the id is missing, this is a static announcement which can not be modified                                           |
| type                 | announcement_type       |                                                                                                                         |
| visibility           | announcement_visibility |                                                                                                                         |
| text                 | +string                 |                                                                                                                         |
| is_currently_visible | ?!boolean               | Not present on static announcements. If false, this announcement currently cannot be seen by non-moderators             |
| has_expired          | ?!boolean               | Not present on static announcements. If true, this announcement has expired. Never true, if `is_currently_visible=true` |

#### announcement_configuration

| Field  | Type                                 | Notes |
|--------|--------------------------------------|-------|
| fields | array of [field_value](#field_value) |       |

#### homepage

| Field             | Type                           | Notes                                                                                                                                                          |
|-------------------|--------------------------------|----------------------------------------------------------------------------------------------------------------------------------------------------------------|
| featured          | array of [featured](#featured) | Order of array is display order (First element at the top)                                                                                                     | 
| upcoming_lectures | array of [lecture](#lecture)   | Does not include chapters and media_sources. Soonest is first element                                                                                          |
| new_lectures      | array of [lecture](#lecture)   | Does not include chapters and media_sources. Latest is first element                                                                                           |
| courses_context   | map of [course](#course)       | Key is id of course. This contains all courses which are referenced in _new_lectures_, _upcoming_lectures_, _featured[...].lecture_ or _featured[...].courses_ |

#### featured

Panels on the homepage showing text or featuring some lectures

| Field            | Type                 | Notes                                                                                                                                                             |
|------------------|----------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| id               | !int                 |                                                                                                                                                                   |
| type             | featured_type        |                                                                                                                                                                   |
| title            | +string              |                                                                                                                                                                   |
| text             | +string              |                                                                                                                                                                   |
| image_url        | ?url                 | If `type=image`                                                                                                                                                   |
| lecture          | ?[lecture](#lecture) | If `type=lecture`. May be null (Lecture has been deleted/does not exist). Does not include chapters and media_sources                                             |
| courses          | ?array of int        | If `type=course_list`. Id's of courses. Courses are in the `courses_context` of [homepage](#homepage)                                                             |
| is_visible       | +!boolean            |                                                                                                                                                                   | 
| display_priority | +!int                | Unique, smallest value is at the top. If updated and another feature already has the same priority as the updated value, this is placed *under* the other feature | 

#### featured_configuration

| Field  | Type                                 | Notes                                                                  |
|--------|--------------------------------------|------------------------------------------------------------------------|
| fields | array of [field_value](#field_value) |                                                                        |

#### course

| Field                          | Type                          | Notes                                                                                                                                                                |
|--------------------------------|-------------------------------|----------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| id                             | int                           | unique                                                                                                                                                               |
| id_string                      | id_string                     | unique                                                                                                                                                               |
| full_name                      | +string                       |                                                                                                                                                                      |
| short_name                     | +string                       |                                                                                                                                                                      |
| semester                       | +?semester_string             |                                                                                                                                                                      |
| organizer                      | +string                       | Name of the organizer                                                                                                                                                |
| topic                          | string                        |                                                                                                                                                                      |
| description                    | +string                       |                                                                                                                                                                      |
| show_chapters_on_course        | boolean                       | If true, chapters should be shown on the course page                                                                                                                 |
| default_authentication_methods | authentication_methods        | Default authentication methods for lectures in this course. Some Lectures might have different methods (A lecture object always contains its authentication methods) |
| authentication_information     | ?string                       | May contain additional information on how the user can authenticate themselves                                                                                       |
| lectures                       | ?array of [lecture](#lecture) | Only present if requested. All (public) lectures. Lectures will include chapters and media_sources                                                                   |
| is_listed                      | !boolean                      | If false, course can only be visited with its url (Unless user is moderator)                                                                                         |
| is_visible                     | !boolean                      | If false, course is only accessible by moderators.                                                                                                                   |

#### course_configuration

| Field       | Type                                    | Notes |
|-------------|-----------------------------------------|-------|
| fields      | array of [field_value](#field_value)    |       |

#### lecture

| Field                  | Type                                    | Notes                                                                           |
|------------------------|-----------------------------------------|---------------------------------------------------------------------------------|
| id                     | int                                     | unique                                                                          |
| course_id              | int                                     |                                                                                 |
| title                  | +string                                 |                                                                                 |
| speaker                | +string                                 | Name of the speaker                                                             |
| location               | +string                                 |                                                                                 |
| time                   | +datetime                               |                                                                                 |
| duration               | +duration                               |                                                                                 |
| description            | +string                                 |                                                                                 |
| thumbnail_url          | ?url                                    |                                                                                 |
| no_recording           | +boolean                                |                                                                                 |
| livestream_planned     | boolean                                 | Always true if a stream is running                                              |
| livestream_url         | ?url                                    | Only present while stream is running                                            |
| allow_embed            | boolean                                 | If true, the lecture may be embedded                                            |
| chapters               | ?array of [chapter](#chapter)           | Only present if requested. Sorted according to the start_time (ascending order) |
| media_sources          | ?array of [media_source](#media_source) | Only present if requested. All (public) media. May be empty                     |
| authentication_methods | authentication_methods                  |                                                                                 |
| is_visible             | +!boolean                               |                                                                                 |
| internal_comment       | +!string                                |                                                                                 |

#### lecture_configuration

| Field       | Type                                 | Notes |
|-------------|--------------------------------------|-------|
| fields      | array of [field_value](#field_value) |       |

#### view_permissions

| Field                | Type                  | Notes                                                                                     |
|----------------------|-----------------------|-------------------------------------------------------------------------------------------|
| type                 | view_permissions_type | For type `authentication` at least one authentication method must be allowed/specified    |
| rwth_authentication  | boolean               | If `type=authentication`                                                                  |
| fsmpi_authentication | boolean               | If `type=authentication`                                                                  |
| moodle_course_ids    | array of int          | If `type=authentication`                                                                  |
| passwords            | map of string         | If `type=authentication`. Key is username. Value is password. Maximum string length: 1024 |

#### chapter

| Field      | Type       | Notes |
|------------|------------|-------|
| id         | !int       |       |
| start_time | +timestamp |       |
| name       | +string    |       |
| is_visible | +!boolean  |       |

#### media_source

video or audio file

| Field            | Type                            | Notes                                 |
|------------------|---------------------------------|---------------------------------------|
| id               | !int                            |                                       |
| quality          | [media_quality](#media_quality) |                                       |
| size             | data_size                       |                                       |
| comment          | +string                         |                                       |
| player_url       | url                             |                                       |
| download_url     | ?url                            | Only present if download is permitted |
| is_visible       | +!boolean                       |                                       |
| internal_comment | +!string                        |                                       |

#### media_quality

| Field        | Type   | Notes                                           |
|--------------|--------|-------------------------------------------------|
| name         | string | Human readable name                             |
| resolution   | string | e.g. `1920x1080` for video or `48kHz` for audio |
| aspect_ratio | string |                                                 |
| priority     | int    |                                                 |

#### feedback_entry

| Field        | Type     | Notes |
|--------------|----------|-------|
| id           | int      |       |
| time_created | datetime |       |
| email        | ?string  |       |
| text         | string   |       |

#### user_info

| Field     | Type   | Notes               |
|-----------|--------|---------------------|
| id        | int    |                     |
| id_string | string |                     |
| name      | string | Human readable name |

#### field_description

<table>
    <thead>
        <th><code>type</code></th>
        <th>Field</th>
        <th>Field Type</th>
        <th>Notes</th>
    </thead>
    <tbody>
        <tr>
            <td rowspan="3">any</td>
            <td>id</td>
            <td>id_string</td>
            <td>Lang keys: <code>object.{id}.name</code> and <code>object.{id}.tooltip</code> (Tooltip may not be present)</td>
        </tr>
        <tr>
            <td>type</td>
            <td>id_string</td>
            <td>As specified in <a href="#Types">Types</a></td>
        </tr>
        <tr>
            <td>default_value</td>
            <td>?according to <code>type</code></td>
            <td>If present, this is a default value which may be used for this field (Field may be omitted at creation then)</td>
        </tr>
        <tr>
            <td rowspan="2"><code>int</code></td>
            <td>min_value</td>
            <td>?int</td>
            <td>inclusive</td>
        </tr>
        <tr>
            <td>max_value</td>
            <td>?int</td>
            <td>inclusive, &gt= <code>min_value</code></td>
        </tr>
        <tr>
            <td rowspan="2"><code>long</code></td>
            <td>min_value</td>
            <td>?long</td>
            <td>inclusive</td>
        </tr>
        <tr>
            <td>max_value</td>
            <td>?long</td>
            <td>inclusive, &gt= <code>min_value</code></td>
        </tr>
        <tr>
            <td rowspan="2"><code>string</code>, <code>url</code></td>
            <td>min_length</td>
            <td>?int</td>
            <td>inclusive, &gt= 0</td>
        </tr>
        <tr>
            <td>max_length</td>
            <td>?int</td>
            <td>inclusive, &gt= <code>min_length</code></td>
        </tr>
        <tr>
            <td rowspan="1"><code>datetime</code>, <code>lecture_id</code>, <code>user_id</code></td>
            <td>may_be_null</td>
            <td>?boolean</td>
            <td>If not present, value may <b>not</b> be null</td>
        </tr>
    </tbody>
</table>

#### field_value

| Field       | Type                                    | Notes |
|-------------|-----------------------------------------|-------|
| description | [field_description](#field_description) |       |
| value       | According to `description.type`         |       |

#### changelog_entry

<table>
    <thead>
        <th><code>type</code></th>
        <th>Field</th>
        <th>Field Type</th>
        <th>Notes</th>
    </thead>
    <tbody>
        <tr>
            <td rowspan="4">any</td>
            <td>id</td>
            <td>int</td>
            <td>ID of this change</td>
        </tr>
        <tr>
            <td>type</td>
            <td>id_string</td>
            <td>Type of this entry. For possible values see this table</td>
        </tr>
        <tr>
            <td>change_time</td>
            <td>datetime</td>
            <td></td>
        </tr>
        <tr>
            <td>modifying_user_id</td>
            <td>int</td>
            <td>ID of user who changed this, may not exist anymore</td>
        </tr>
        <tr>
            <td rowspan="5"><code>modification</code></td>
            <td>object_type</td>
            <td>id_string</td>
            <td></td>
        </tr>
        <tr>
            <td>field_description</td>
            <td><a href="#field_description">field_description</a></td>
            <td></td>
        </tr>
        <tr>
            <td>object_id</td>
            <td>int</td>
            <td>ID of the object which was modified</td>
        </tr>
        <tr>
            <td>old_value</td>
            <td>?according to <code>field.type</code></td>
            <td>Not present if this is at creation</td>
        </tr>
        <tr>
            <td>new_value</td>
            <td>according to <code>field.type</code></td>
            <td></td>
        </tr>
        <tr>
            <td rowspan="3"><code>deletion_change</code></td>
            <td>object_type</td>
            <td>id_string</td>
            <td></td>
        </tr>
        <tr>
            <td>object_id</td>
            <td>int</td>
            <td>ID of the object which was modified</td>
        </tr>
        <tr>
            <td>is_now_deleted</td>
            <td>boolean</td>
            <td></td>
        </tr>
        <tr>
            <td rowspan="5"><code>creation</code></td>
            <td>object_type</td>
            <td>id_string</td>
            <td></td>
        </tr>
        <tr>
            <td>object_id</td>
            <td>int</td>
            <td>ID of the object which was created</td>
        </tr>
        <tr>
            <td>variant</td>
            <td>?id_string</td>
            <td>ID of variant as which the object was created. Only present if this object type has a variant</td>
        </tr>
        <tr>
            <td>parent_type</td>
            <td>?id_string</td>
            <td>Type of the parent object. Only present if this object has a parent</td>
        </tr>
        <tr>
            <td>parent_id</td>
            <td>?int</td>
            <td>ID of the parent object. Only present if this object has a parent</td>
        </tr>
        <tr>
            <td rowspan="5"><code>unknown</code></td>
            <td>unknown_object_type</td>
            <td>string</td>
            <td></td>
        </tr>
        <tr>
            <td>unknown_field</td>
            <td>string</td>
            <td></td>
        </tr>
        <tr>
            <td>object_id</td>
            <td>int</td>
            <td>ID of the object which was modified</td>
        </tr>
        <tr>
            <td>param1</td>
            <td>string</td>
            <td>May be null</td>
        </tr>
        <tr>
            <td>param2</td>
            <td>string</td>
            <td></td>
        </tr>
    </tbody>
</table>

#### sorter_log_entry

| Field            | Type      | Notes                                                                                                                                 |
|------------------|-----------|---------------------------------------------------------------------------------------------------------------------------------------|
| time             | datetime  |                                                                                                                                       |
| file_path        | string    |                                                                                                                                       |
| media_source_id  | int       |                                                                                                                                       |
| lecture_id       | ?int      | Present if, and only if, the media source is not deleted (Note that it may be implicitly deleted if the lecture or course is deleted) |
| lecture_title    | ?string   | Same as `lecture_id`                                                                                                                  |
| lecture_time     | ?datetime | Same as `lecture_id`                                                                                                                  |
| course_id        | ?int      | Same as `lecture_id`                                                                                                                  |
| course_full_name | ?string   | Same as `lecture_id`                                                                                                                  |

#### sorter_error_log_entry

| Field                | Type          | Notes |
|----------------------|---------------|-------|
| id                   | int           |       |
| time                 | datetime      |       |
| file_path            | string        |       |
| matched_course_id    | ?int          |       |
| matched_lectures_ids | ?array of int |       |

#### settings

| Field    | Type                                           | Notes |
|----------|------------------------------------------------|-------|
| sections | array of [settings_section](#settings_section) |       |

#### settings_section

| Field   | Type                                       | Notes                              |
|---------|--------------------------------------------|------------------------------------|
| name    | id_string                                  | Lang Key: `setting.section.{name}` |
| entries | array of [settings_entry](#settings_entry) |                                    |

#### settings_entry

<table>
    <thead>
        <th><code>type</code></th>
        <th>Field</th>
        <th>Field Type</th>
        <th>Notes</th>
    </thead>
    <tbody>
        <tr>
            <td rowspan="2">any</td>
            <td>id</td>
            <td>id_string</td>
            <td>Lang Key: <code>setting.entry.{id}.name</code> and <code>setting.entry.{id}.description</code> (description may not exist)</td>
        </tr>
        <tr>
            <td>type</td>
            <td>id_string</td>
            <td>Type of this entry. For possible values see this table</td>
        </tr>
        <tr>
            <td rowspan="1"><code>subsection</code></td>
            <td>entries</td>
            <td>array of settings_entry</td>
            <td>Contain no subsections</td>
        </tr>
        <tr>
            <td rowspan="1"><code>boolean</code></td>
            <td>value</td>
            <td>boolean</td>
            <td></td>
        </tr>
    </tbody>
</table>

#### job

| Field            | Type       | Notes                                                                |
|------------------|------------|----------------------------------------------------------------------|
| id               | int        |                                                                      |
| type             | id_string  | This is not job_type since the database may contain legacy job types |
| priority         | id_string  |                                                                      |
| worker_id        | ?id_string |                                                                      |
| last_worker_ping | ?datetime  |                                                                      |
| state            | id_string  |                                                                      |
| is_canceled      | boolean    |                                                                      |
| creation_time    | datetime   |                                                                      |
| start_time       | ?datetime  |                                                                      |
| finish_time      | ?datetime  |                                                                      |
| data             | string     |                                                                      |
| status           | string     |                                                                      |

#### worker

| Field     | Type     | Notes  |
|-----------|----------|--------|
| hostname  | string   | Unique |
| last_ping | datetime |        |

#### api_error

| Field      | Type      | Notes                            |
|------------|-----------|----------------------------------|
| error_code | id_string | Possible values are listed below |
| message    | string    | Human readable message           |

Possible `error_code`:

| Code                                  | HTTP Code | Notes                                                |
|---------------------------------------|-----------|------------------------------------------------------|
| bad_request                           | 400       |                                                      |
| method_not_allowed                    | 405       |                                                      |
| malformed_json                        | 400       | Content-Type needs to be set to `application/json`   |
| request_missing_parameter             | 400       |                                                      |
| request_invalid_parameter             | 400       |                                                      |
| unknown_request_path                  | 400       |                                                      |
| internal_server_error                 | 500       |                                                      |
| lecture_has_no_password               | 400       |                                                      |
| authentication_failed                 | 403       |                                                      |
| authentication_not_available          | 503       |                                                      |
| unauthorized                          | 401       |                                                      |
| csrf_token_invalid                    | 403       | CSRF token must be included as `X-Csrf-Token` Header |
| access_forbidden                      | 403       |                                                      |
| rate_limited                          | 429       |                                                      |
| unknown_object                        | 404       |                                                      |
| object_error                          | 400       |                                                      |
| modification_unexpected_current_value | 409       |                                                      |
| too_many_suggestions                  | 403       |                                                      |
| site_is_readonly                      | 503       |                                                      |
| site_is_disabled                      | 503       |                                                      |
| site_is_overloaded                    | 503       |                                                      |

## Changelog

### v0.54

* Added `PUT /feedback/new`, `GET /feedback?entries_per_page={entries_per_page}&page={page}`
* Added `feedback_entry` object

### v0.53

* Changed type of `course_id_string` to regex. Added forbidden values.

### v0.52

* Added error code `site_is_overloaded`

### v0.51

* Added `server_name` to `GET /status`

### v0.50

* Removed `GET /announcements`
* Updated `GET /status`. Among other things, this now returns the announcements
* Changed name of error type `authentication_servers_not_available` to `authentication_not_available`
* Added error types `site_is_readonly` and `site_is_disabled`
* Added `GET /is_running`

### v0.49

* Added `GET /status`

### v0.48

* Removed `PUT /job/new/{type}`
* Removed `POST /sorter/run`

### v0.47

* Removed `GET /object_management/{object_type}/{object_id}/{field_id}/changelog`
* Removed `GET /object_management/{object_type}/{object_id}/changelog`
* Added query parameters to `GET /object_management/changelog` instead
* Update type `unknown` of `changelog_entry`:
  * `unknown_type` -> `unknown_object_type` 
  * `old_value` -> `param1` (May be null)
  * `new_value` -> `param2` 

### v0.46

* Type `url` may be type of field.

### v0.45

* Removed `permission` object
* Removed `permissions` from the configurations of courses and lectures
* Added `view_permissions` object which may appear as the type of some field
* `may_be_null` may now only appear in `field_description` of type `datetime`, `lecture_id` or `user_id`
* `duration` and `timestamp` may no longer appear as the type of some field

### v0.44

* Added `expected_current_values` to `PATCH /object_management/{object_type}/{object_id}/configuration` (and accordingly to `POST /object_management/modify_many`)
* Added `modification_unexpected_current_value` error code

### v0.43

* Added `parent_type` and `parent_id` to `changelog_entry` of type `creation`

### v0.42

* Added `worker` object. Added `workers` field to `GET /jobs`
* Removed `POST /job/{id}/restart`, `POST /job/{id}/cancel` and `POST /job/{id}/copy`
* Added `POST /job/{id}/action` instead

### v0.41

* Added `GET /jobs`, `PUT /job/new/{type}`, `DELETE /job/{id}`, `POST /job/{id}/restart`, `POST /job/{id}/cancel` and `POST /job/{id}/copy`
* Added `job` object
* Added `job_state` and `job_type` types

### v0.40

* Added `GET /user/me/settings` and `PATCH /user/me/settings`
* Added `settings`, `settings_section` and `settings_entry` object

### v0.39

* Added `sorter_log_entry` and `sorter_error_log_entry` objects
* Added `GET /sorter/log`, `DELETE /sorter/error_log/{log_id}` and `POST /sorter/run`

### v0.38

* Removed `is_modifiable` of `field_description` (All fields can be modified)
* Fields with default values may be omitted at creation now
* Added guarantees for the creation of some objects

### v0.37

* CSRF token is now returned in a cookie instead of the response body 

### v0.36

* Added CSRF tokens. The token is returned by `POST /authentication/status`.
* CSRF token is required by:
  - `GET /lecture_import/rwth_online/{rwth_course_id}`
  - `PATCH /object_management/{object_type}/{object_id}/configuration`
  - `POST /object_management/modify_many`
  - `PUT /object_management/{object_type}/new`
  - `DELETE /object_management/{object_type}/{object_id}`
  - `POST /object_management/{object_type}/{object_id}/resurrect`
* Added error code `csrf_token_invalid`

### v0.35

* Added `POST /object_management/{object_type}/{object_id}/resurrect`

### v0.34

* Added error code `rate_limited`

### v0.33

* Removed `field_entry`
* Added `field_description` and `field_value` instead. The format of the configuration endpoints and `changelog_entry` has changed accordingly. 

### v0.32

* Changes to [changelog_entry](#changelog_entry) format
* Added `GET /object_management/{object_type}/{object_id}/changelog?entries_per_page={entries_per_page}&page={page}`

### v0.31

* Changed `GET /{object_type}/{object_id}/configuration` to `GET /object_management/{object_type}/{object_id}/configuration`
* Changed `PATCH /{object_type}/{object_id}/configuration` to `PATCH /object_management/{object_type}/{object_id}/configuration`
* Changed `POST /objects/modify_many` to `POST /object_management/modify_many`
* Changed `GET /{object_type}/new/configuration` to `GET /object_management/{object_type}/new/configuration`
* Changed `PUT /{object_type}/new/create` to `PUT /object_management/{object_type}/new`
* Changed `DELETE /{object_type}/{object_id}` to `DELETE /object_management/{object_type}/{object_id}`
* Changed `GET /{object_type}/{object_id}/{field_id}/changelog` to `GET /object_management/{object_type}/{object_id}/{field_id}/changelog`
* Changed `GET /objects/changelog` to `GET /object_management/changelog`

### v0.30

* Updated `GET /{object_type}/{object_id}/{field_id}/changelog`
* Added `GET /search?q={search_term}`

### v0.29

* Added `GET /{object_type}/{object_id}/{field_id}/changelog`
* Changed `/lecture_import/rwthonline/{rwth_course_id}` to ``/lecture_import/rwth_online/{rwth_course_id}``

### v0.28

* `/lecture_import/list_document/{course_id}` => `/lecture_import/rwthonline/{rwth_course_id}`
* Removed `POST /lecture_import/calendar_export`

### v0.27

* Added `GET /lecture_import/list_document/{course_id}`
* Added `POST /lecture_import/calendar_export`

### v0.26

* Added `GET /objects/changelog?entries_per_page={entries_per_page}&page={page}`

### v0.25

* Allowed creation of `chapter` objects
* Added `PUT /lecture/{lecture_id}/chapter_suggestion`
* Added `too_many_suggestions` error type

### v0.24

* Added `allow_embed` to `lecture`

### v0.23

* Updated directly modifiable Fields. They are now indicated by a `+`
* Added `legacy_announcement_type`

### v0.22

* Changed featured type `text` to `plain` (Because all types contain text)

### v0.21

* Updated `display_priority` of `featured` to be unique and no longer nullable

### v0.20

* Added `DELETE /{object_type}/{object_id}`

### v0.19

* Changed type of `courses` in `featured` to array of int. Added courses to the `courses_context` of the homepage instead.

### v0.18

* Renamed Object Types to Object Variants as this was confusing. Renamed `type_fields` to `variant_fields` and `type` to `variant` for this

### v0.17

* Added `POST /objects/modify_many`

### v0.16

* Added `display_priority` to `featured` (Also to [Directly Modifiable Fields](#Directly-Modifiable-Fields))

### v0.15

* Added `course_filter_criteria` type
* `id_string` may no longer be type of field entry
* Added `course_id_string` type

### v0.14

* Added `GET /users`

### v0.13

* Reverted `GET /announcement` back to `GET /announcements`
* Reverted `GET /course` back to `GET /courses`

### v0.12

* Added `POST /authentication/moderator_logout`

### v0.11

* Added Object Modification API
* Added `announcement_configuration`, `featured_configuration`, `course_configuration`, `lecture_configuration`, `permission` and `field_entry` objects
* Added `internal_comment` to `media_source`
* Added `unknown_object` and `object_error` error types. Removed `unknown_course` and `unknown_lecture` error types.
* Changed `GET /announcements` to `GET /announcement` (for consistency)
* Changed `GET /courses` to `GET /course` (for consistency)
* Changed `GET /courses/{course_id}` to `GET /course/{course_id}` (for consistency)
* Changed `GET /lectures/{lecture_id}` to `GET /lecture/{lecture_id}` (for consistency)

### v0.10

* Added `id`, `is_currently_visible` and `has_expired` to `announcement` object
* Added `id` and `is_visible` to `featured` object
* Added `is_listed` and `is_visible` to `course` object
* Added `is_visible` and `internal_comment` to `lecture` object
* Added `id` and `is_visible` to `chapter` object
* Added `id` and `is_visible` to `media_source` object
* Changed pattern for `id_string` from `[a-z0-9_-]{1,100}` to `(?=.*[a-z_-])[a-z0-9_-]{1,100}`
* Changed route `GET /courses/{course_id}/lectures/{lecture_id}` to `GET /lectures/{lecture_id}`
* Made `lecture` of `featured` nullable 

### v0.9

* Added `user_info` to response of `POST /authentication/status`

### v0.8

* Added `authentication_information` to `course` object

### v0.7

* Added `bad_request` and `method_not_allowed` error codes

### v0.6

* Changed `/authentication/start` to `/authentication/start_oauth`
* Removed `/authentication/login`
* Added `/authentication/password` and `/authentication/fsmpi` instead

### v0.5

* Changed `moderator` authentication type to `fsmpi` (To be consistent with `authentication_methods`).

### v0.3

* `GET /courses/{course_id}/lectures/{lecture_id}` now accepts course id_string and includes the course

### v0.2

* `GET /courses` returns map instead of array

### v0.1

* Initial version
