from typing import Callable

from psycopg.errors import Diagnostic, SerializationFailure
from psycopg import Connection, Cursor
from psycopg.cursor import TUPLES_OK
import psycopg

from api.database import DbValueType, DbResultSet, FilledStatement
from api.database.database import PreparedStatement, DbConnectionFactory, DatabaseWarning, DatabaseError, \
    DbAffectedRows, \
    DatabaseResultRow, TransactionIsolationLevel
from api.database.abstract_py_connector import PythonDbConnection

print(f"Using psycopg (Postgres) Connector Version: {psycopg.version.__version__}")


class _PostgresDbConnection(PythonDbConnection[Connection, Cursor]):
    
    def __init__(self,
                 use_multi_query: bool,
                 fetch_warnings: bool,
                 create_new_connection: Callable[[], Connection],
                 py_connection: Connection):
        super().__init__(
            "%s",
            (),  # Can't continue transaction after any exceptions
            py_connection)
        self._use_multi_query = use_multi_query
        if fetch_warnings:
            self._py_connection.add_notice_handler(self._handle_notice)
        self._get_warnings = fetch_warnings
        self._create_new_connection = create_new_connection
        self._disconnected = False
        self._last_warnings: list[DatabaseWarning] = []
    
    def _handle_notice(self, diagnostic: Diagnostic):
        self._last_warnings.append(DatabaseWarning(f"""\
severity: {diagnostic.severity}
severity_nonlocalized: {diagnostic.severity_nonlocalized}
sqlstate: {diagnostic.sqlstate}
message_primary: {diagnostic.message_primary}
message_detail: {diagnostic.message_detail}
message_hint: {diagnostic.message_hint}
statement_position: {diagnostic.statement_position}
internal_position: {diagnostic.internal_position}
internal_query: {diagnostic.internal_query}
context: {diagnostic.context}
schema_name: {diagnostic.schema_name}
table_name: {diagnostic.table_name}
column_name: {diagnostic.column_name}
datatype_name: {diagnostic.datatype_name}
constraint_name: {diagnostic.constraint_name}
source_file: {diagnostic.source_file}
source_line: {diagnostic.source_line}
source_function: {diagnostic.source_function}
"""))
    
    def _can_use_multi_query(self, statements: [FilledStatement]):
        return self._use_multi_query and all(map(lambda stat: len(stat.values) == 0, statements))
    
    def is_disconnected(self, force_ping: bool = False) -> bool:
        if self._closed:
            raise RuntimeError("Already closed")
        if self._disconnected:
            return True
        if not self._py_connection.closed:  # There is no ping method
            return False
        self._disconnected = True
        return True
    
    def try_reconnect(self) -> bool:
        # noinspection PyBroadException
        try:
            new_conn = self._create_new_connection()
        except Exception:
            return False
        # noinspection PyBroadException
        try:
            self._py_connection.close()
        except Exception:
            pass
        self._clear_cache_on_new_connection()
        self._py_connection = new_conn
        if self._get_warnings:
            self._py_connection.add_notice_handler(self._handle_notice)
        self._disconnected = False
        return True

    def close(self):
        self._closed = True
        self._py_connection.close()
    
    def get_transaction_begin_statements(self, writable: bool, isolation_level: TransactionIsolationLevel) -> list[PreparedStatement or str]:
        if writable:
            return [f"START TRANSACTION ISOLATION LEVEL {isolation_level.value}, READ WRITE"]
        else:
            return [f"START TRANSACTION ISOLATION LEVEL {isolation_level.value}, READ ONLY"]
    
    def get_transaction_end_statement(self, commit: bool) -> PreparedStatement or str:
        if commit:
            return "COMMIT"
        else:
            return "ROLLBACK"
    
    def _create_cursor(self, prepared: bool):
        return self._py_connection.cursor()
    
    def _is_transaction_conflict_exception(self, exception: Exception) -> bool:
        return isinstance(exception, SerializationFailure)
    
    def _db_execute_single_statement(self,
                                     cursor: Cursor,
                                     statement: str,
                                     values: list[DbValueType]) -> tuple[list[DatabaseWarning], DbAffectedRows, DbResultSet]:
        cursor.execute(statement.encode(encoding="UTF-8"), params=values)
        all_rows = self._fetch_all_rows(cursor)
        if len(all_rows) > 0:
            column_mapping = PythonDbConnection._create_column_mapping(cursor)
        return (self._fetch_warnings(),
                cursor.rowcount,
                list(map(lambda row: DatabaseResultRow(column_mapping, tuple(row)), all_rows)))
    
    def _db_execute_multiple_statements(self,
                                        cursor: Cursor,
                                        statements: str,
                                        values: list[DbValueType]) -> list[tuple[list[DatabaseWarning], DbAffectedRows, DbResultSet]]:
        cursor.execute(statements.encode(encoding="UTF-8"), params=values, prepare=False)
        results = []
        warnings = self._fetch_warnings()
        while True:
            all_rows = self._fetch_all_rows(cursor)
            if len(all_rows) > 0:
                column_mapping = PythonDbConnection._create_column_mapping(cursor)
            results.append((warnings,
                            cursor.rowcount,
                            list(map(
                                lambda row: DatabaseResultRow(column_mapping, tuple(row)),
                                all_rows
                            ))))
            if not cursor.nextset():
                break
        return results
    
    @staticmethod
    def _fetch_all_rows(cursor: Cursor) -> list[tuple]:
        if not cursor.pgresult or cursor.pgresult.status != TUPLES_OK:
            return []
        return cursor.fetchall()
    
    def _fetch_warnings(self) -> list[DatabaseWarning]:
        if not self._get_warnings:
            return []
        warnings = self._last_warnings
        self._last_warnings = []
        return warnings
    
    def _db_execute_script(self, cursor: Cursor, script: str):
        cursor.execute(script.encode(encoding="UTF-8"), prepare=False)


class PostgresDbConnectionFactory(DbConnectionFactory):
    
    def __init__(self, use_multi_query: bool, fetch_warnings: bool,
                 host: str, port: int,
                 user: str, password: str,
                 database: str):
        super().__init__()
        self._use_multi_query = use_multi_query
        self._fetch_warnings = fetch_warnings
        self._host = host
        self._port = port
        self._user = user
        self._password = password
        self._database = database
    
    def supports_per_transaction_writeable_flag(self) -> bool:
        return True
    
    def _create_new_connection(self) -> Connection:
        return Connection.connect(
            host=self._host,
            port=self._port,
            user=self._user,
            password=self._password,
            dbname=self._database,
            autocommit=True  # Yep, this is correct. When DISABLED, BEGIN's are send automatically, and we don't want that
        )
    
    def new_connection(self, writable: bool = True) -> _PostgresDbConnection:
        try:
            new_conn = self._create_new_connection()
        except Exception as e:
            raise DatabaseError("An exception occurred while connecting to database") from e
        return _PostgresDbConnection(self._use_multi_query, self._fetch_warnings, self._create_new_connection, new_conn)
