from api.database import *
from api.miscellaneous import *
from api.course import lecture_queue_query_auth, course_queue_query_auth, course_query_auth
from api.authentication import ViewPermissions, ViewPermissionsType, get_effective_view_permissions

from api.objects.changelog_entries import ModificationEntry, DeletionEntry, CreationEntry
from api.objects.field import ObjectColumnField, ObjectIndirectField, InterimValueType
from api.objects.type import ValueType


# Cannot import due to circular dependency
# noinspection PyUnresolvedReferences
class SimpleObjectColumnField(ObjectColumnField):
    
    def __init__(self,
                 api_id: str,
                 include_in_config: bool,
                 type: ValueType,
                 db_column: str,
                 default_value_json: JsonTypes or tuple[None] or None = None):
        super().__init__(api_id, include_in_config, type.api_id, db_column, default_value_json)
        self._type = type
    
    def _add_type_restrictions_to(self, config_entry):
        self._type.add_restrictions_to(config_entry)
    
    def get_current_value_from_db(self, object_db: DbResultRow):
        return self._type.db_value_to_json(object_db[self.db_column])
    
    def do_value_update(self,
                        transaction: AbstractTransaction,
                        changelog_update: "ChangelogUpdate",
                        object_id: int,
                        current_object_db: DbResultRow,
                        expected_old_client_value: CJsonValue or None,
                        new_client_value: CJsonValue) -> tuple[str, DbValueType] or None:
        new_value_db = self._type.validate_and_convert_client_value(transaction, object_id, new_client_value)
        new_value_json = self._type.db_value_to_json(new_value_db)
        old_value_json = self._type.db_value_to_json(current_object_db[self.db_column])
        if expected_old_client_value is not None and not expected_old_client_value.equals_json(old_value_json):
            raise ApiClientException(ERROR_MODIFICATION_UNEXPECTED_CURRENT_VALUE)
        if new_value_json == old_value_json:
            return None
        changelog_update.add_entry(ModificationEntry(
            self,
            object_id,
            old_value_json,
            new_value_json
        ))
        return self.db_column, new_value_db
    
    def do_value_creation(self,
                          transaction: AbstractTransaction,
                          changelog_update: "CreationChangelogUpdate",
                          new_client_value: CJsonValue) -> tuple[str, DbValueType]:
        new_value_db = self._type.validate_and_convert_client_value(transaction, None, new_client_value)
        new_value_json = self._type.db_value_to_json(new_value_db)
        changelog_update.add_object_creation_entry(lambda object_id: ModificationEntry(
            self,
            object_id,
            None,
            new_value_json,
            True
        ))
        return self.db_column, new_value_db


# Cannot import due to circular dependency
# noinspection PyUnresolvedReferences
class ListField(ObjectIndirectField):
    
    def __init__(self,
                 api_id: str,
                 include_in_config: bool,
                 type_api_id: str,
                 entry_type: ValueType,
                 list_table: str,
                 list_id_column: str,
                 list_entry_column: str,
                 default_value_json: JsonTypes or tuple[None] or None = None):
        super().__init__(api_id, include_in_config, type_api_id, default_value_json)
        self._entry_type = entry_type
        self._list_table = db_ensure_valid_id(list_table)
        self._list_id_column = db_ensure_valid_id(list_id_column)
        self._list_entry_column = db_ensure_valid_id(list_entry_column)
        
        self._query_statement = PreparedStatement(f"""
            SELECT "{list_entry_column}" FROM "{list_table}"
            WHERE "{list_id_column}" = ?
            """)
    
    def _add_type_restrictions_to(self, config_entry):
        pass
    
    def queue_get_current_statements(self, transaction: ReadTransaction, object_id: int) -> tuple[FilledStatement, ...]:
        return (transaction.queue_statement(self._query_statement, object_id), )
    
    def get_json_value_from_results(self,
                                    transaction: ReadTransaction,
                                    object_db: DbResultRow,
                                    statements: tuple[FilledStatement, ...]) -> JsonTypes:
        return list(map(lambda row: self._entry_type.db_value_to_json(row[self._list_entry_column]),
                        transaction.get_result(statements[0])))
    
    def validate_and_convert_client_value(self,
                                          transaction: AbstractTransaction,
                                          object_id: int or None,
                                          client_value: CJsonValue) -> InterimValueType:
        old_list_db = None
        if object_id is not None:
            current_data_db = transaction.execute_statement(self._query_statement, object_id)
            old_list_db = list(map(lambda row: row[self._list_entry_column], current_data_db))
        
        new_list_db = list(map(lambda c: self._entry_type.validate_and_convert_client_value(
                                                transaction, object_id, c),
                               client_value.as_array()))
        return old_list_db, new_list_db
    
    def queue_update(self,
                     transaction: WriteTransaction,
                     changelog_update: "ChangelogUpdate",
                     object_id: int,
                     expected_old_client_value: CJsonValue or None,
                     list_data: InterimValueType):
        old_list_db, new_list_db = list_data
        old_list_json = list(map(self._entry_type.db_value_to_json, old_list_db))
        new_list_json = list(map(self._entry_type.db_value_to_json, new_list_db))
        
        if expected_old_client_value is not None and not expected_old_client_value.equals_json(old_list_json):
            raise ApiClientException(ERROR_MODIFICATION_UNEXPECTED_CURRENT_VALUE)
        
        if new_list_json == old_list_json:
            return None
        
        if len(new_list_db) == 0:
            transaction.queue_statement(f"""
                        DELETE FROM "{self._list_table}"
                        WHERE "{self._list_id_column}" = ?
                        """, object_id)
        else:
            transaction.queue_statement(f"""
                        INSERT INTO "{self._list_table}" ("{self._list_id_column}", "{self._list_entry_column}")
                        WITH "new_data_2095705247" ("new_value_2095705247") AS (
                            VALUES {",".join(map(lambda _: "(?)", range(0, len(new_list_db))))}
                        )
                        SELECT ?, "new_data_2095705247"."new_value_2095705247"
                        FROM "new_data_2095705247"
                        LEFT JOIN "{self._list_table}"
                        ON ("{self._list_table}"."{self._list_id_column}" = ?
                            AND "{self._list_table}"."{self._list_entry_column}" = "new_data_2095705247"."new_value_2095705247")
                        WHERE ("{self._list_table}"."{self._list_entry_column}" is NULL)
                        """, *new_list_db, object_id, object_id)
            transaction.queue_statement(f"""
                        DELETE FROM "{self._list_table}"
                        WHERE "{self._list_id_column}" = ?
                            AND "{self._list_entry_column}" NOT IN (
                                VALUES {",".join(map(lambda _: "(?)", range(0, len(new_list_db))))}
                            )
                        """, object_id, *new_list_db)
        
        changelog_update.add_entry(ModificationEntry(
            self,
            object_id,
            old_list_json,
            new_list_json
        ))
    
    def queue_creation(self,
                       transaction: WriteTransaction,
                       changelog_update: "CreationChangelogUpdate",
                       object_id: int,
                       list_data: InterimValueType):
        old_list_db, new_list_db = list_data
        new_list_json = list(map(self._entry_type.db_value_to_json, new_list_db))
        if len(new_list_db) > 0:
            transaction.queue_statement(f"""
                        INSERT INTO "{self._list_table}" ("{self._list_id_column}", "{self._list_entry_column}")
                        VALUES {",".join(map(lambda _: f"(?, ?)", range(0, len(new_list_db))))}
                        """, *[v for ele in new_list_db for v in [object_id, ele]])
        changelog_update.add_entry(ModificationEntry(
            self,
            object_id,
            None,
            new_list_json,
            True
        ))


# Cannot import due to circular dependency
# noinspection PyUnresolvedReferences
class FeaturedDisplayPriorityField(ObjectIndirectField[tuple[int, DbResultSet]]):
    
    def __init__(self):
        super().__init__(
            "display_priority",
            False,
            "int")
        
        self._query_all_statement = PreparedStatement("""
            SELECT *
            FROM "featured"
            WHERE (NOT "deleted")
            ORDER BY CASE WHEN "order" IS NULL THEN 0 ELSE "order" END ASC, "id" ASC
            """)
        self._update_one_statement = PreparedStatement(f"""
            UPDATE "featured"
            SET "order" = ?
            WHERE "id" = ?
            """)
    
    def _add_type_restrictions_to(self, config_entry):
        pass
    
    def queue_get_current_statements(self, transaction: ReadTransaction, object_id: int) -> tuple[FilledStatement, ...]:
        return ()
    
    def get_json_value_from_results(self,
                                    transaction: ReadTransaction,
                                    object_db: DbResultRow,
                                    statements: tuple[FilledStatement, ...]) -> JsonTypes:
        return object_db["order"]
    
    def validate_and_convert_client_value(self,
                                          transaction: AbstractTransaction,
                                          object_id: int or None,
                                          client_value: CJsonValue) -> InterimValueType:
        return (client_value.as_int(MIN_VALUE_SINT32, MAX_VALUE_SINT32),
                transaction.execute_statement(self._query_all_statement))
    
    def queue_update(self,
                     transaction: WriteTransaction,
                     changelog_update: "ChangelogUpdate",
                     object_id: int,
                     expected_old_client_value: CJsonValue or None,
                     new_value: InterimValueType):
        new_prio, featured_list_db = new_value
        current_prio = 0
        current_order = None
        for feature_db in featured_list_db:
            if feature_db["id"] != object_id:
                current_prio += 1
                continue
            if new_prio == current_prio:
                return
            if new_prio > current_prio:
                new_prio -= 1
            current_order = FeaturedDisplayPriorityField._get_order(feature_db)
            featured_list_db = featured_list_db[:current_prio] + featured_list_db[(current_prio + 1):]
            break
        
        if expected_old_client_value is not None and not expected_old_client_value.equals_json(current_prio):
            raise ApiClientException(ERROR_MODIFICATION_UNEXPECTED_CURRENT_VALUE)
        
        new_object_order = self._update_priorities_on_change(transaction, changelog_update, featured_list_db, new_prio)
        if new_object_order == current_order:
            return
        transaction.queue_statement(self._update_one_statement, new_object_order, object_id)
        changelog_update.add_entry(ModificationEntry(
            self,
            object_id,
            current_order,
            new_object_order  # TODO
        ))
    
    def queue_creation(self,
                       transaction: WriteTransaction,
                       changelog_update: "CreationChangelogUpdate",
                       object_id: int,
                       new_value: InterimValueType):
        new_prio, featured_list_db = new_value
        object_order = self._update_priorities_on_change(transaction, changelog_update, featured_list_db, new_prio)
        transaction.queue_statement(f"""
                    UPDATE "featured"
                    SET "order" = ?
                    WHERE "id" = ?
                    """, object_order, object_id)
        changelog_update.add_entry(ModificationEntry(
            self,
            object_id,
            None,
            new_list_json,
            True
        ))
    
    @staticmethod
    def _get_order(featured_db):
        order = featured_db["order"]
        if order is None:
            return 0
        return order
    
    def _update_priorities_on_change(
            self,
            transaction: WriteTransaction,
            changelog_update: "ChangelogUpdate",
            feature_list_db: DbResultSet,
            new_prio: int,) -> int:
        """

        :param feature_list_db: Must not contain the entry for the object itself
        :param new_prio:
        :return: First int is new prio for object, list tuple: id, old prio, new prio
        """
        if len(feature_list_db) == 0:
            return 0
        if new_prio < 0:
            first_order = FeaturedDisplayPriorityField._get_order(feature_list_db[0])
            return first_order - 1
        if new_prio >= len(feature_list_db):
            last_order = FeaturedDisplayPriorityField._get_order(feature_list_db[-1])
            return last_order + 1
        
        target_order = FeaturedDisplayPriorityField._get_order(feature_list_db[new_prio])
        
        if new_prio >= 1 and FeaturedDisplayPriorityField._get_order(feature_list_db[new_prio - 1]) < target_order - 1:
            # We can just decrease order of feature at new_prio by one
            changelog_update.add_entry(ModificationEntry(
                self,
                feature_list_db[new_prio]["id"],
                feature_list_db[new_prio]["order"],
                target_order - 1
            ))
            transaction.queue_statement(self._update_one_statement, target_order - 1, feature_list_db[new_prio]["id"])
            return target_order
        
        # Put feature after current feature at new_prio
        target_order += 1
        feature_order = target_order
        
        #  Move all features down until unused order number is found
        i = new_prio + 1
        while i < len(feature_list_db):
            feature_db = feature_list_db[i]
            if FeaturedDisplayPriorityField._get_order(feature_db) > target_order:
                break
            target_order += 1

            changelog_update.add_entry(ModificationEntry(
                self,
                feature_db["id"],
                feature_db["order"],
                target_order
            ))
            transaction.queue_statement(self._update_one_statement, target_order, feature_db["id"])
            
            i += 1
        
        return feature_order


_SQL_GET_LECTURE_AUTH = PreparedStatement("""
SELECT * FROM "perm"
WHERE "perm"."lecture_id" = ?
    AND (NOT "deleted")
""")
_SQL_CREATE_COURSE_PERM = PreparedStatement("""
INSERT INTO "perm" ("course_id", "type", "param1", "param2")
    VALUES (?, ?, ?, ?)
RETURNING "id"
""")
_SQL_CREATE_LECTURE_PERM = PreparedStatement("""
INSERT INTO "perm" ("lecture_id", "type", "param1", "param2")
    VALUES (?, ?, ?, ?)
RETURNING "id"
""")
_SQL_DELETE_PERM = PreparedStatement("""
UPDATE "perm"
SET "deleted" = true
WHERE "id" = ?
""")


# Cannot import due to circular dependency
# noinspection PyUnresolvedReferences
class ViewPermissionsField(ObjectIndirectField[tuple[DbResultSet, ViewPermissions]]):
    
    def __init__(self,
                 api_id: str,
                 include_in_config: bool,
                 for_lecture: bool):
        """
        
        :param api_id:
        :param include_in_config:
        :param for_lecture: If true, this is for a lecture, otherwise it is for a course
        """
        super().__init__(
            api_id,
            include_in_config,
            "view_permissions",
            ViewPermissions(ViewPermissionsType.INHERIT).to_json())
        self._for_lecture = for_lecture
    
    def _add_type_restrictions_to(self, config_entry):
        pass
    
    def queue_get_current_statements(self, transaction: ReadTransaction, object_id: int) -> tuple[FilledStatement, ...]:
        if self._for_lecture:
            return (lecture_queue_query_auth(transaction, object_id),)
        return (course_queue_query_auth(transaction, object_id),)
    
    def get_json_value_from_results(self,
                                    transaction: ReadTransaction,
                                    object_db: DbResultRow,
                                    statements: tuple[FilledStatement, ...]) -> JsonTypes:
        return get_effective_view_permissions(
            transaction.get_result(statements[0]),
            self._for_lecture).object_permissions.to_json()
    
    def validate_and_convert_client_value(self,
                                          transaction: AbstractTransaction,
                                          object_id: int or None,
                                          client_value: CJsonValue) -> tuple[DbResultSet, ViewPermissions]:
        return (
            None if object_id is None
            else (transaction.execute_statement(_SQL_GET_LECTURE_AUTH, object_id) if self._for_lecture
                  else course_query_auth(object_id, transaction)),
            ViewPermissions.from_json(client_value))
    
    def _create_perm(self,
                     transaction: WriteTransaction,
                     changelog_update: "ChangelogUpdate",
                     object_id: int,
                     type: str,
                     param1: str or None,
                     param2: str or None):
        perm_id = transaction.execute_statement(
            _SQL_CREATE_LECTURE_PERM if self._for_lecture else _SQL_CREATE_COURSE_PERM,
            object_id,
            type,
            param1,
            param2)[0]["id"]
        from api.objects.objects import CLASS_COURSE, CLASS_LECTURE, CLASS_PERMISSION
        from api.objects.types import TYPE_PERMISSION_TYPE
        changelog_update.add_entry(CreationEntry(CLASS_PERMISSION.get_creation_subject(),
                                                 perm_id,
                                                 TYPE_PERMISSION_TYPE.db_value_to_json(type),
                                                 (CLASS_LECTURE if self._for_lecture else CLASS_COURSE),
                                                 object_id))
        if type == "moodle":
            changelog_update.add_entry(ModificationEntry(
                CLASS_PERMISSION.get_field("moodle_course_id"),
                perm_id,
                None,
                param1,
                True
            ))
        if type == "password":
            changelog_update.add_entry(ModificationEntry(
                CLASS_PERMISSION.get_field("username"),
                perm_id,
                None,
                param1,
                True
            ))
            changelog_update.add_entry(ModificationEntry(
                CLASS_PERMISSION.get_field("password"),
                perm_id,
                None,
                param2,
                True
            ))
    
    # noinspection PyMethodMayBeStatic
    def _delete_perm(self,
                     transaction: WriteTransaction,
                     changelog_update: "ChangelogUpdate",
                     perm_id: int):
        transaction.queue_statement(_SQL_DELETE_PERM, perm_id)
        from api.objects.objects import CLASS_PERMISSION
        changelog_update.add_entry(DeletionEntry(CLASS_PERMISSION.get_deletion_subject(), perm_id, True))
    
    def queue_update(self,
                     transaction: WriteTransaction,
                     changelog_update: "ChangelogUpdate",
                     object_id: int,
                     expected_old_client_value: CJsonValue or None,
                     new_value: tuple[DbResultSet, ViewPermissions]):
        old_perm_list_db, view_perm = new_value
        if expected_old_client_value is not None \
                and not expected_old_client_value.equals_json(
                    get_effective_view_permissions(old_perm_list_db, self._for_lecture).object_permissions.to_json()
                ):
            raise ApiClientException(ERROR_MODIFICATION_UNEXPECTED_CURRENT_VALUE)
        
        match view_perm.type:
            case ViewPermissionsType.INHERIT:
                for old_perm_db in old_perm_list_db:
                    self._delete_perm(transaction, changelog_update, old_perm_db["id"])
                
            case ViewPermissionsType.PUBLIC:
                has_public = False
                for old_perm_db in old_perm_list_db:
                    if old_perm_db["type"] == "public":
                        has_public = True
                        continue
                    self._delete_perm(transaction, changelog_update, old_perm_db["id"])
                if not has_public:
                    self._create_perm(transaction, changelog_update, object_id, "public", None, None)
                
            case ViewPermissionsType.PRIVATE:
                has_none = False
                for old_perm_db in old_perm_list_db:
                    if old_perm_db["type"] == "none":
                        has_none = True
                        continue
                    self._delete_perm(transaction, changelog_update, old_perm_db["id"])
                if not has_none:
                    self._create_perm(transaction, changelog_update, object_id, "none", None, None)
                
            case ViewPermissionsType.AUTHENTICATION:
                make_rwth = view_perm.rwth_authentication
                make_fsmpi = view_perm.fsmpi_authentication
                remaining_moodle_course_ids: list[int] = view_perm.moodle_course_ids.copy()
                remaining_passwords: dict[str, str] = view_perm.passwords.copy()
                
                for old_perm_db in old_perm_list_db:
                    match old_perm_db["type"]:
                        case "rwth":
                            if not make_rwth:
                                self._delete_perm(transaction, changelog_update, old_perm_db["id"])
                            make_rwth = False
                        case "fsmpi":
                            if not make_fsmpi:
                                self._delete_perm(transaction, changelog_update, old_perm_db["id"])
                            make_fsmpi = False
                        case "moodle":
                            try:
                                moodle_id = int(old_perm_db["param1"])
                            except ValueError:
                                self._delete_perm(transaction, changelog_update, old_perm_db["id"])
                                continue
                            if moodle_id in remaining_moodle_course_ids:
                                remaining_moodle_course_ids.remove(moodle_id)
                            else:
                                self._delete_perm(transaction, changelog_update, old_perm_db["id"])
                        case "password":
                            username = old_perm_db["param1"]
                            password = old_perm_db["param2"]
                            if username not in remaining_passwords or password != remaining_passwords[username]:
                                self._delete_perm(transaction, changelog_update, old_perm_db["id"])
                            else:
                                remaining_passwords.pop(username)
                        case "public" | "none" | "l2p" | _:
                            self._delete_perm(transaction, changelog_update, old_perm_db["id"])
                
                if make_rwth:
                    self._create_perm(transaction, changelog_update, object_id, "rwth", None, None)
                if make_fsmpi:
                    self._create_perm(transaction, changelog_update, object_id, "fsmpi", None, None)
                
                for moodle_id in remaining_moodle_course_ids:
                    self._create_perm(transaction, changelog_update, object_id, "moodle", str(moodle_id), None)
                
                for username, password in remaining_passwords.items():
                    self._create_perm(transaction, changelog_update, object_id, "password", username, password)
            case _:
                raise ValueError(f"Unknown view permission type {perm.type}")  # pragma: no cover
    
    def queue_creation(self,
                       transaction: WriteTransaction,
                       changelog_update: "CreationChangelogUpdate",
                       object_id: int,
                       new_value: tuple[DbResultSet, ViewPermissions]):
        _, perm = new_value
        match perm.type:
            case ViewPermissionsType.INHERIT:
                pass
            case ViewPermissionsType.PUBLIC:
                self._create_perm(transaction, changelog_update, object_id, "public", None, None)
            case ViewPermissionsType.PRIVATE:
                self._create_perm(transaction, changelog_update, object_id, "none", None, None)
            case ViewPermissionsType.AUTHENTICATION:
                if perm.rwth_authentication:
                    self._create_perm(transaction, changelog_update, object_id, "rwth", None, None)
                if perm.fsmpi_authentication:
                    self._create_perm(transaction, changelog_update, object_id, "fsmpi", None, None)
                
                for moodle_id in perm.moodle_course_ids:
                    self._create_perm(transaction, changelog_update, object_id, "moodle", str(moodle_id), None)
                
                for username, password in perm.passwords.items():
                    self._create_perm(transaction, changelog_update, object_id, "password", username, password)
            case _:
                raise ValueError(f"Unknown view permission type {perm.type}")  # pragma: no cover
